@@ -1,5 +1,6 @@
 # SPDX-License-Identifier: BSD-3-Clause
 import asyncio
+import functools
 import logging
 import socket
 import ssl
@@ -47,6 +48,8 @@ from .types import PayloadType, T
 MQTT_LOGGER = logging.getLogger("mqtt")
 MQTT_LOGGER.setLevel(logging.WARNING)
 
+T = TypeVar("T")
+
 
 class ProtocolVersion(IntEnum):
     """
@@ -105,6 +108,7 @@ class Client:
         properties: Optional[Properties] = None,
         message_retry_set: int = 20,
         socket_options: Optional[Iterable[SocketOption]] = (),
+        max_concurrent_outgoing_calls: Optional[int] = None,
     ):
         self._hostname = hostname
         self._port = port
@@ -123,6 +127,14 @@ class Client:
         self._pending_calls_threshold: int = 10
         self._misc_task: Optional["asyncio.Task[None]"] = None
 
+        self._outgoing_calls_sem: Optional[asyncio.Semaphore]
+        if max_concurrent_outgoing_calls is not None:
+            self._outgoing_calls_sem = asyncio.Semaphore(
+                max_concurrent_outgoing_calls
+            )
+        else:
+            self._outgoing_calls_sem = None
+
         if protocol is None:
             protocol = ProtocolVersion.V311
 
@@ -218,6 +230,21 @@ class Client:
         if not self._disconnected.done():
             self._disconnected.set_result(None)
 
+    # TODO: Simplify the logic that surrounds `self._outgoing_calls_sem` with
+    # `nullcontext` when we support Python 3.10 (`nullcontext` becomes async-aware in
+    # 3.10). See: https://docs.python.org/3/library/contextlib.html#contextlib.nullcontext
+    def _outgoing_call(self, func: Callable[..., Awaitable[T]]) -> Callable[..., Awaitable[T]]:
+        @functools.wraps(func)
+        async def decorated(*args: Any, **kwargs: Any) -> T:
+            if not self._outgoing_calls_sem:
+                return await func(*args, **kwargs)
+
+            async with self._outgoing_calls_sem:
+                return await func(*args, **kwargs)
+
+        return decorated
+
+    @_outgoing_call
     async def subscribe(self, *args: Any, timeout: int = 10, **kwargs: Any) -> int:
         result, mid = self._client.subscribe(*args, **kwargs)
         # Early out on error
@@ -229,6 +256,7 @@ class Client:
             # Wait for cb_result
             return await self._wait_for(cb_result, timeout=timeout)
 
+    @_outgoing_call
     async def unsubscribe(self, *args: Any, timeout: int = 10) -> None:
         result, mid = self._client.unsubscribe(*args)
         # Early out on error
@@ -240,6 +268,7 @@ class Client:
             # Wait for confirmation
             await self._wait_for(confirmation.wait(), timeout=timeout)
 
+    @_outgoing_call
     async def publish(self, *args: Any, timeout: int = 10, **kwargs: Any) -> None:
         info = self._client.publish(*args, **kwargs)  # [2]
         # Early out on error
